mrGearboxMogli = {};

function mrGearboxMogli.prerequisitesPresent(specializations) 
	return SpecializationUtil.hasSpecialization(RealisticMotorized, specializations);
end;


function mrGearboxMogli:load(xmlFile) 

	--print(" ----- ******  loading mrGearboxMogli ****** ------");

	self.isRealistic = true;	
	
	--you have to select the next transmission mode here (example : if only one additionnal transmission, the modeNum should be 4)
	--! don't forget to "order" your transmission specialization in the moddesc.xml vehicleType section (T4 first, then T5)
	self.clutch = mrGearboxMogli.clutch;
	self.setNextGear = mrGearboxMogli.setNextGear;
	self.setNextRange = mrGearboxMogli.setNextRange;
	self.setGear = mrGearboxMogli.setGear;
	self.getClutchPercent = mrGearboxMogli.getClutchPercent;
	
	self.doGearShift = mrGearboxMogli.doGearShift;
	
	self.setGearboxMogliModeOn = mrGearboxMogli.setGearboxMogliModeOn;
	self.setGearboxMogliModeOff = mrGearboxMogli.setGearboxMogliModeOff;
	
	self.setManualGear3 = RealisticMotorized.setManualGear3;
	
	
	local xmlString = "vehicle.gearboxMogli";
	
	self.mrGbMClutchRpmRatio          = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchMinRpm"), 1.1 * self.realEngine.idleRpm) / self.realEngine.ratedRpm; 
	self.mrGbMClutchEngagingTime      = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchEngagingTimeMs"), 2000); -- 1000 = 1s
	self.mrGbMTransmissionEfficiency  = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#transmissionEfficiency"), self.realTransmissionEfficiency);
	self.mrGbMTransmissionEfficiencyBackup = self.realTransmissionEfficiency;
		
	self.mrGbMGearTimeToShiftGear     = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#shiftTimeMs"), 1000);
	self.mrGbMGearTimeToShiftHl       = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges#shiftTimeMs"), 1000);
	self.mrGbMGearTimeToShiftReverse  = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#shiftTimeMs"), 1000);

	self.mrGbMClutchAfterShiftGear    = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#clutchRatio"), 0.3);
	self.mrGbMClutchAfterShiftHl      = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges#clutchRatio"), 0.3);
	self.mrGbMClutchAfterShiftReverse = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#clutchRatio"), 0.1);
	
	self.mrGbMGears = {};
	local i = 0;
	while true do
		local baseName = xmlString .. string.format(".gears.gear(%d)", i);		
		local speed    = getXMLFloat(xmlFile, baseName .. "#speed");
		if speed==nil then
			break;
		end;
		i = i + 1;
		local name  = Utils.getNoNil(getXMLString(xmlFile, baseName .. "#name"), tostring(i));
		local newGear = {speed=speed/3.6,name=name};
		table.insert(self.mrGbMGears, newGear); -- m/s
	end;	
	
	self.mrGbMRanges = {};
	local i = 0;
	while true do
		local baseName = xmlString .. string.format(".ranges.range(%d)", i);		
		local ratio = getXMLFloat(xmlFile, baseName .. "#ratio");
		if ratio==nil then
			break;
		end;
		i = i + 1;
		local name     = Utils.getNoNil(getXMLString(xmlFile, baseName .. "#name"), "G"..tostring(i));
		local newRange = {ratio=ratio,name=name};
		table.insert(self.mrGbMRanges, newRange); -- m/s
	end;	
	
	self.mrGbMReverseRatio            = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#ratio"), 1);
	self.mrGbMReverseResetGear        = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".reverse#resetGear"), true);
	self.mrGbMReverseResetRange       = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".reverse#resetRange"), true);
	self.mrGbMReverseMinGear          = math.max( Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".reverse#minGear"), 1), 1);
	self.mrGbMReverseMaxGear          = math.min( Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".reverse#maxGear"), table.getn(self.mrGbMGears)), table.getn(self.mrGbMGears));
	self.mrGbMReverseMinRange         = math.max( Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".reverse#minRange"), 1), 1);
	self.mrGbMReverseMaxRange         = math.min( Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".reverse#maxRange"), table.getn(self.mrGbMRanges)), table.getn(self.mrGbMRanges));
	
	--RealisticUtils.testClass("mrGearboxMogli - mrGbMGears", self.mrGbMGears);
	
	self.mrGbMDefaultGear     = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".gears#defaultGear"), 1);
	self.mrGbMDefaultGear0    = self.mrGbMDefaultGear;
	self.mrGbMDefaultRange    = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".ranges#defaultRange"), table.getn(self.mrGbMRanges));
	self.mrGbMDefaultRange0   = self.mrGbMDefaultRange;
	self.mrGbMRangeGearOffset = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".ranges#gearOffset"), 0);

	self.mrGbMAutoShiftTimeout   = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#autoShiftTimeout"), 4000 );
	self.mrGbMAutoShiftUpTimer   = 0
	self.mrGbMAutoShiftDownTimer = 0
	
	local f
	f = getXMLFloat(xmlFile, xmlString .. ".gears#autoUpRpm");
	if f ~= nil then
		self.mrGbMAutoShiftUpRpmRatio = f / self.realEngine.ratedRpm
	end
	f = getXMLFloat(xmlFile, xmlString .. ".gears#autoDownRpm");
	if f ~= nil then
		self.mrGbMAutoShiftDownRpmRatio = f / self.realEngine.ratedRpm
	end
	
	self.mrGbMIsOn = false;
	
	self.mrGbMCurrentGear  = self.mrGbMDefaultGear;
	self.mrGbMCurrentRange = self.mrGbMDefaultRange; -- 1=forward low range, 2 = forward high range, -1 = reverse
	self.mrGbMMinGear      = 1; -- 1 = first gear
	self.mrGbMMaxGear      = table.getn(self.mrGbMGears); -- to be set each time we switch from reverse/forward
	self.mrGbMMinRange     = 1; -- 1 = first gear
	self.mrGbMMaxRange     = table.getn(self.mrGbMRanges); -- to be set each time we switch from reverse/forward
	
	self.mrGbMGearShiftingNeeded = false;
	self.mrGbMGearShiftingTime = 0;
	
	self.mrGbMReverseActive = false;	
	self.mrGbMNeutralActive = true;	
	self.mrGbMClutchPressed = false;
	self.mrGbMAutomatic     = self.mrGbMAutoShiftUpRpmRatio ~= nil or self.mrGbMAutoShiftDownRpmRatio ~= nil;
	
	--adding the new transmission mode
	self.realTransmissionMode.maxMode = self.realTransmissionMode.maxMode + 1
	self.mrGbMTransmissionModeNum     = self.realTransmissionMode.maxMode
	self.realTransmissionMode.modes[self.mrGbMTransmissionModeNum] = {};
	self.realTransmissionMode.modes[self.mrGbMTransmissionModeNum].enabled = true; -- gearboxMogli mode
	self.realTransmissionMode.modes[self.mrGbMTransmissionModeNum].text = "manual";	
	
end;

function mrGearboxMogli:delete()    
end;

function mrGearboxMogli:mouseEvent(posX, posY, isDown, isUp, button)
end;

function mrGearboxMogli:keyEvent(unicode, sym, modifier, isDown)
end;

function mrGearboxMogli:readStream(streamId, connection)
	self.mrGbMIsOn = streamReadBool(streamId);
	self.mrGbMCurrentGear = streamReadInt32(streamId);
	self.mrGbMCurrentRange = streamReadInt32(streamId);
	self.mrGbMReverseActive = streamReadBool(streamId);
end;
 
function mrGearboxMogli:writeStream(streamId, connection)
	streamWriteBool(streamId, self.mrGbMIsOn);	
	streamWriteInt32(streamId, self.mrGbMCurrentGear);
	streamWriteInt32(streamId, self.mrGbMCurrentRange);
	streamWriteBool(streamId, self.mrGbMReverseActive);	
end;

function mrGearboxMogli:updateTick(dt)

	if self.isAITractorActivated and self.mrGbMIsOn then
		self:setGearboxMogliModeOff();	
	else -- not ai controlled
		if self.realIsMotorStarted then
		
			--new powershift mode = T4
			if      self.motor.speedLevel == 0 
					and self.realTransmissionMode.currentMode==self.mrGbMTransmissionModeNum then
				if not self.mrGbMIsOn then
					self:setGearboxMogliModeOn();	
				end;				
			else		
				if self.mrGbMIsOn then
					self:setGearboxMogliModeOff();
				end;				
			end;
			
		end;
	end;
	
end;

function mrGearboxMogli:update(dt)


	if self:getIsActiveForInput(false) then
	
		if self.mrGbMIsOn and self.realIsMotorStarted then		
			if g_gui.currentGui == nil and (self.isMouseActive == nil or not self.isMouseActive) then
				--print(self.time .. " mrGearboxMogli:update - g_gui.currentGuiName = " .. tostring(g_gui.currentGuiName));				
				if InputBinding.hasEvent(InputBinding.mrGearboxMogliNEUTRAL) then
					if self.mrGbMAutoShiftUpRpmRatio ~= nil or self.mrGbMAutoShiftDownRpmRatio ~= nil then
						if self.mrGbMNeutralActive == self.mrGbMAutomatic then
							self:setGear(self.mrGbMCurrentGear, self.mrGbMDefaultRange, self.mrGbMReverseActive, not self.mrGbMNeutralActive);
							self.mrGbMAutomatic = true
						else
							self.mrGbMAutomatic = not self.mrGbMAutomatic 
						end
					
						if self.mrGbMAutomatic then
							self.realTransmissionMode.modes[self.mrGbMTransmissionModeNum].text = "auto";	
						else
							self.realTransmissionMode.modes[self.mrGbMTransmissionModeNum].text = "manual";	
						end
					else	
						self:setGear(self.mrGbMCurrentGear, self.mrGbMDefaultRange, self.mrGbMReverseActive, not self.mrGbMNeutralActive);
					end
				elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliREVERSE) then			
					self:setGear(self.mrGbMCurrentGear, self.mrGbMCurrentRange, not self.mrGbMReverseActive, self.mrGbMNeutralActive);					
				elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGEUP) then -- high/low range shift
					self:setNextRange(false);                                       
				elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGEDOWN) then -- high/low range shift
					self:setNextRange(true);
				elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTGEARUP) then
					self:setNextGear(false);
				elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTGEARDOWN) then
					self:setNextGear(true);	
				end;
				local clutchPressed = InputBinding.isPressed(InputBinding.mrGearboxMogliCLUTCH)
				if self.mrGbMClutchPressed ~= clutchPressed then
					self:clutch(clutchPressed);
				end;
			end;			
		end;

		
	end;
	
	
	if self.isServer and self.isActive and self.mrGbMIsOn then
		if self.mrGbMmotorLoad == nil then
			self.mrGbMmotorLoad = math.min( math.max( self.realMotorLoad, 0), 1)
		else
			
			local f
			if self.realMotorLoad < self.mrGbMmotorLoad then
				f = math.min( math.max( dt * 0.025, 0.1 ), 1 )
			else
				f = math.min( math.max( dt * 0.00025, 0.001 ), 1 )
			end
			self.mrGbMmotorLoad = self.mrGbMmotorLoad + f * ( math.min( math.max( self.realMotorLoad, 0), 1) - self.mrGbMmotorLoad )
		end
	
		self.realParkbrakeIsManual = true;
		self.realParkbrakeIsActive = false;
		self.realClutchEngaged     = true;				
		
		if self.mrGbMGearShiftingNeeded then
			self.realClutchEngaged = false;
			if self.time>=self.mrGbMGearShiftingTime then
				self.mrGbMGearShiftingNeeded = false;
				self:doGearShift();
			else
				self.lastAcceleration      = math.min(0.25,self.lastAcceleration)
				self.realParkbrakeIsManual = false;				
			end;
		else
			local clutchDelta = -dt/self.mrGbMClutchEngagingTime
			if self.mrGbMNeutralActive then
				self.realParkbrakeIsActive = true;
				self.realClutchEngaged     = false;	
				self.realClutchPercent     = math.max(0.1, self.realClutchPercent -dt/self.mrGbMClutchEngagingTime );
			else
				--no auto parkbrake when a gear is engaged

				local idleRpmRatio     = self.mrGbMClutchRpmRatio
				if idleRpmRatio > self.realEngine.idleRpmRatio then
					local base = self.realEngine.idleRpmRatio
					local diff = idleRpmRatio - base
					idleRpmRatio = self.mrGbMmotorLoad * diff + base
				elseif idleRpmRatio < self.realEngine.idleRpmRatio then
					idleRpmRatio = self.realEngine.idleRpmRatio
				end
				
				-- semi automatic transmission
				if      self.mrGbMAutomatic
						and ( self.realClutchPercent        > 0.5
						   or self.mrGbMCurrentGear         > self.mrGbMDefaultGear0 )
						and not self.mrGbMClutchPressed then
										
					-- shift down
					if self.mrGbMAutoShiftDownRpmRatio ~= nil and self.mrGbMCurrentGear > self.mrGbMMinGear then
						local base = idleRpmRatio
						local diff = self.mrGbMAutoShiftDownRpmRatio - base
						local autoDownRpmRatio  = self.mrGbMmotorLoad * diff + base						
						local ratio = self.mrGbMRanges[self.mrGbMCurrentRange].ratio * self.realEngine.idleRpmRatio;
						if self.mrGbMReverseActive then	
							ratio = ratio * self.mrGbMReverseRatio;
						end 
						local speed = self.mrGbMGears[self.mrGbMCurrentGear].speed * ratio;

						if math.abs(self.realGroundSpeed) < speed and self.mrGbMCurrentGear > self.mrGbMDefaultGear0 then
							for i=self.mrGbMCurrentGear-1,self.mrGbMDefaultGear0,-1 do
								speed = self.mrGbMGears[i].speed * ratio;
								if self.mrGbMReverseActive then	
									speed = speed * self.mrGbMReverseRatio;
								end 
								if math.abs(self.realGroundSpeed) >= speed then
									self.mrGbMAutoShiftDownTimer = 0
									self.mrGbMAutoShiftUpTimer   = 0 --self.time + self.mrGbMAutoShiftTimeout 
									self:setGear(i, self.mrGbMCurrentRange, self.mrGbMReverseActive, self.mrGbMNeutralActive);
									break;
								end
							end
						elseif  self.time                       > self.mrGbMAutoShiftDownTimer
								and self.realMotorThrottleForWheels > 0.1
								and self.realSoundEngineRevFx       < autoDownRpmRatio then
							self:setNextGear(true);
							clutchDelta = 0
						end
					end
					
					-- shift up
					if self.mrGbMAutoShiftUpRpmRatio ~= nil and self.mrGbMCurrentGear < self.mrGbMMaxGear then				
						local base = Utils.getNoNil( self.mrGbMAutoShiftDownRpmRatio, idleRpmRatio )
						local diff = self.mrGbMAutoShiftUpRpmRatio - base
						local autoUpRpmRatio  = self.mrGbMmotorLoad * diff + base
						local ratio = self.mrGbMRanges[self.mrGbMCurrentRange].ratio * math.max(1,self.mrGbMAutoShiftUpRpmRatio)
						if self.mrGbMReverseActive then	
							ratio = ratio * self.mrGbMReverseRatio;
						end 
						local speed = self.mrGbMGears[self.mrGbMCurrentGear].speed * ratio;
						
						if math.abs(self.realGroundSpeed) > speed then
										
							for i=self.mrGbMCurrentGear+1,self.mrGbMMaxGear,1 do
								speed = self.mrGbMGears[i].speed * ratio;
								if math.abs(self.realGroundSpeed) <= speed then
									self.mrGbMAutoShiftDownTimer = 0 --self.time + self.mrGbMAutoShiftTimeout 
									self.mrGbMAutoShiftUpTimer   = 0
									self:setGear(i, self.mrGbMCurrentRange, self.mrGbMReverseActive, self.mrGbMNeutralActive);
									break;
								end
							end
							
						elseif  self.time                       > self.mrGbMAutoShiftUpTimer 
								and self.realMotorThrottleForWheels > 0.1
								and self.realClutchPercent          > 0.9
								and self.realSoundEngineRevFx       > autoUpRpmRatio then
							self:setNextGear(false);
						end
					end
				end

			--print(tostring(self.mrGbMmotorLoad).." "..tostring(idleRpmRatio).." "..tostring(autoDownRpmRatio).." "..tostring(autoUpRpmRatio).." "..tostring(self.mrGbMAutoShiftUpTimer-self.time))
				
				local targetClutchPercent = self:getClutchPercent( self.realEngine.currentGearRatedSpeed, idleRpmRatio );
				
				if targetClutchPercent < self.realClutchPercent then
					self.realClutchPercent = math.max(0.1, self.realClutchPercent - math.min( dt/self.mrGbMClutchEngagingTime, self.realClutchPercent - targetClutchPercent ) );
				else
					self.realClutchPercent = math.min(  1, self.realClutchPercent + dt/self.mrGbMClutchEngagingTime );
				end;
			end;
		end;


		--[[test
		if self.realTestValue==0 then
			self.realParkbrakeIsManual = true;
			self.realParkbrakeIsActive = false;
			self.realClutchEngaged = false;	
		end;]]
		
	end;

end;

function mrGearboxMogli:getClutchPercent( gearSpeed, minRpmRatio )

	if self.mrGbMClutchPressed then
		return 0.1
	end
	
	local currentRpmRatio = math.max(0, self.realAvgMotorizedWheelSpd ) / gearSpeed;
	local targetRpmRatio  = math.max( minRpmRatio, currentRpmRatio );
	local idleRpmRatio    = math.max(self.realEngine.idleRpmRatio, self.lastAcceleration);
	
	if minRpmRatio <= currentRpmRatio then
		clutchPercent = 1
	elseif currentRpmRatio < idleRpmRatio + 0.001 then
		clutchPercent = 1 - targetRpmRatio / idleRpmRatio
	else
		clutchPercent = ( targetRpmRatio - idleRpmRatio ) / ( currentRpmRatio - idleRpmRatio );
	end
	
	if clutchPercent < 0.1 then
		clutchPercent = 0.1
	elseif clutchPercent > 1 then
		clutchPercent = 1
	end
	
	return clutchPercent;
end;

function mrGearboxMogli:draw() 	
	
	if self.mrGbMIsOn then
		setTextBold(false);
		setTextColor(1, 1, 1, 1);
		setTextAlignment(RenderText.ALIGN_LEFT);
			
		local gearText = self.mrGbMRanges[self.mrGbMCurrentRange].name .." ".. self.mrGbMGears[self.mrGbMCurrentGear].name
		if self.mrGbMNeutralActive then
			gearText = gearText .. " (N)";
		elseif self.mrGbMReverseActive then
			gearText = gearText .. " (R)";
		end;
		
		local rpm = 0;
		
		--if self.isServer then
		--	rpm = self.mrGbMMaxRpm*self.realAvgMotorizedWheelSpd/self.realManualGearMaxSpeed;
		--else
		rpm = self.realSoundEngineRevFx*self.realEngine.ratedRpm;
		--end;
		
		
		renderText(0.86, 0.70, 0.03, "Gear : " .. gearText);			
		renderText(0.86, 0.67, 0.03, string.format("Rpm  : %4.0f", rpm));

		if self.isServer then
			local clutch = 10 * math.floor( 10 * self.realClutchPercent + 0.5 )
			renderText(0.86, 0.64, 0.03, string.format("Clutch : %3.0f", clutch).."%");
		end;
	end;
	
end;


function mrGearboxMogli:clutch(pressed,noEventSend)
	self.mrGbMClutchPressed = pressed;
	
	if noEventSend == nil or noEventSend == false then
		if g_server == nil then		
			--client side : tell the server what we want to do (gear up or down)
			g_client:getServerConnection():sendEvent(mrGearboxMogliClutchEvent:new(self,pressed));
		end;
	end;
end

function mrGearboxMogli:setNextGear(gearDown, noEventSend)

	
	if self.isServer then
		--server side : compute the new gear
		--print(self.time .. " mrGearboxMogli:setNextGear - gearDown=" .. tostring(gearDown) .. " / self.mrGbMCurrentGear=" .. tostring(self.mrGbMCurrentGear));

		local wantedGear = nil;
		
		if gearDown then
			if self.mrGbMCurrentGear>self.mrGbMMinGear then
				self.mrGbMAutoShiftDownTimer = 0
				self.mrGbMAutoShiftUpTimer   = self.time + self.mrGbMAutoShiftTimeout 
				wantedGear = self.mrGbMCurrentGear - 1;		
			end;
		else
			if self.mrGbMCurrentGear<self.mrGbMMaxGear then
				self.mrGbMAutoShiftDownTimer = self.time + self.mrGbMAutoShiftTimeout 
				self.mrGbMAutoShiftUpTimer   = 0
				wantedGear = self.mrGbMCurrentGear + 1;		
			end;
		end;
		
		if wantedGear~=nil then
			self:setGear(wantedGear, self.mrGbMCurrentRange, self.mrGbMReverseActive, self.mrGbMNeutralActive);
		end;	
	end;	
	
	if noEventSend == nil or noEventSend == false then
		if g_server == nil then		
			--client side : tell the server what we want to do (gear up or down)
			g_client:getServerConnection():sendEvent(mrGearboxMogliSetNextGearEvent:new(self, gearDown));
		end;
	end;

end;

function mrGearboxMogli:setNextRange(rangeDown, noEventSend)

	
	if self.isServer and self.mrGbMCurrentRange~=0 then
		--server side : compute the new gear
		--print(self.time .. " mrGearboxMogli:setNextGear - rangeDown=" .. tostring(rangeDown) .. " / self.mrGbMCurrentGear=" .. tostring(self.mrGbMCurrentGear));

		local wantedRange = nil;
		
		if rangeDown then
			if self.mrGbMCurrentRange>self.mrGbMMinRange then
				wantedRange = self.mrGbMCurrentRange - 1;		
			end;
		else
			if self.mrGbMCurrentRange<self.mrGbMMaxRange then
				wantedRange = self.mrGbMCurrentRange + 1;		
			end;
		end;
		
		if wantedRange~=nil then
			self.mrGbMAutoShiftDownTimer = 0
			self.mrGbMAutoShiftUpTimer   = 0
			self:setGear(self.mrGbMCurrentGear, wantedRange, self.mrGbMReverseActive, self.mrGbMNeutralActive);
		end;	
	end;	
	
	if noEventSend == nil or noEventSend == false then
		if g_server == nil then		
			--client side : tell the server what we want to do (gear up or down)
			g_client:getServerConnection():sendEvent(mrGearboxMogliSetNextRangeEvent:new(self, rangeDown));
		end;
	end;

end;

function mrGearboxMogli:setGear(gearNum, range, reverseActive, neutralActive, noEventSend) -- range : 0 = neutral, 1 = low, 2 = high, -1 = reverse

	--print(self.time .. " mrGearboxMogli - setGear - gearNum=" .. tostring(gearNum) .. " - range=" .. tostring(range));

	
	if self.isServer then		
	
		if reverseActive~=self.mrGbMReverseActive  then
		
			self.mrGbMReverseActive = reverseActive;
			
			--timer to shift the "reverse/forward"
			self.mrGbMGearShiftingNeeded = true;
			self.mrGbMGearShiftingTime   = math.max( self.mrGbMGearShiftingTime, self.time + self.mrGbMGearTimeToShiftReverse );
			self.realClutchPercent       = math.min( self.realClutchPercent, self.mrGbMClutchAfterShiftReverse )

			if reverseActive then
				self.mrGbMMinGear  = self.mrGbMReverseMinGear
				self.mrGbMMaxGear  = self.mrGbMReverseMaxGear
				self.mrGbMMinRange = self.mrGbMReverseMinRange
				self.mrGbMMaxRange = self.mrGbMReverseMaxRange
			else
				self.mrGbMMinGear  = 1
				self.mrGbMMaxGear  = table.getn(self.mrGbMGears);
				self.mrGbMMinRange = 1
				self.mrGbMMaxRange = table.getn(self.mrGbMRanges);
			end;
			
			local default = self.mrGbMCurrentGear
			if self.mrGbMReverseResetGear then
				default = self.mrGbMDefaultGear
			end
			
			self.mrGbMDefaultGear  = self.mrGbMCurrentGear
			self.mrGbMCurrentGear  = math.min( math.max( default, self.mrGbMMinGear ), self.mrGbMMaxGear )
			
			local default = self.mrGbMCurrentRange
			if self.mrGbMReverseResetRange then
				default = self.mrGbMDefaultRange
			end

			self.mrGbMDefaultRange = self.mrGbMCurrentRange
			self.mrGbMCurrentRange = math.min( math.max( default, self.mrGbMMinRange ), self.mrGbMMaxRange )


			self.mrGbMAutoShiftDownTimer = self.time + self.mrGbMAutoShiftTimeout 
			self.mrGbMAutoShiftUpTimer   = self.time + self.mrGbMAutoShiftTimeout 
		elseif neutralActive~=self.mrGbMNeutralActive then
		
			self.mrGbMNeutralActive      = neutralActive
			self.mrGbMGearShiftingNeeded = true;
			self.mrGbMGearShiftingTime   = math.max( self.mrGbMGearShiftingTime, self.time + self.mrGbMGearTimeToShiftHl );
			self.realClutchPercent       = math.min( self.realClutchPercent, self.mrGbMClutchAfterShiftHl )
			
		elseif range~=self.mrGbMCurrentRange then -- range shifting	
		
			newRange = math.min( math.max( range, self.mrGbMMinRange ), self.mrGbMMaxRange )
			if     newRange > self.mrGbMCurrentRange then
				self.mrGbMCurrentGear  = math.min( math.max( self.mrGbMCurrentGear - self.mrGbMRangeGearOffset, self.mrGbMMinGear ), self.mrGbMMaxGear )
			elseif newRange < self.mrGbMCurrentRange then
				self.mrGbMCurrentGear  = math.min( math.max( self.mrGbMCurrentGear + self.mrGbMRangeGearOffset, self.mrGbMMinGear ), self.mrGbMMaxGear )
			end			
			self.mrGbMCurrentRange       = newRange

			--timer to shift the "range"
			self.mrGbMGearShiftingNeeded = true;
			self.mrGbMGearShiftingTime   = math.max( self.mrGbMGearShiftingTime, self.time + self.mrGbMGearTimeToShiftHl );
			self.realClutchPercent       = math.min( self.realClutchPercent, self.mrGbMClutchAfterShiftHl )
			
		else -- gear shifting ?					
			
			self.mrGbMCurrentGear  = math.min( math.max( gearNum, self.mrGbMMinGear ), self.mrGbMMaxGear )
			--timer to set the gear
			self.mrGbMGearShiftingNeeded = true;
			self.mrGbMGearShiftingTime   = math.max( self.mrGbMGearShiftingTime, self.time + self.mrGbMGearTimeToShiftGear );
			self.realClutchPercent       = math.min( self.realClutchPercent, self.mrGbMClutchAfterShiftGear )
		end;		
		
	end;
	
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			-- server side :  send the new "gear" to all the clients
			g_server:broadcastEvent(mrGearboxMogliSetGearEvent:new(self, gearNum, range, reverseActive, neutralActive), nil, nil, self);
		else
			-- client side : tell the server the gearNum we want it to set
			g_client:getServerConnection():sendEvent(mrGearboxMogliSetGearEvent:new(self, gearNum, range, reverseActive, neutralActive));
		end;
	end;

end;




function mrGearboxMogli:setGearboxMogliModeOn()

	--print(self.time .. " mrGearboxMogli:setGearboxMogliModeOn");

	self.mrGbMIsOn = true;
	
	if self.isServer then
		self.realDirectInverser = false;
		self.realShuttleIsManual = true;
		self.realUseDefaultShuttle = false;
		self.realClutchEngaged = true;
		
		self.realTransmissionEfficiency = self.mrGbMTransmissionEfficiency;
			
		if self.realShuttleDirection<0 then		
			self:setGear(self.mrGbMDefaultGear, self.mrGbMDefaultRange, true, self.mrGbMNeutralActive); -- first gear, H range, reverse
		else
			self:setGear(self.mrGbMDefaultGear, self.mrGbMDefaultRange, false, self.mrGbMNeutralActive); -- first gear, H range, forward
		end;
	end;

end;

function mrGearboxMogli:setGearboxMogliModeOff()

	--print(self.time .. " mrGearboxMogli:setGearboxMogliModeOff");

	self.mrGbMIsOn = false;
	
	if self.isServer then
		self.mrGbMDefaultGear  = self.mrGbMCurrentGear
		self.mrGbMDefaultRange = self.mrGbMCurrentRange
		self.realClutchEngaged = true;	
		self:setManualGear3(false);
		self.mrGbMGearShiftingNeeded = false;
		self.realTransmissionEfficiency = self.mrGbMTransmissionEfficiencyBackup;
	end;
	
	self.realParkbrakeIsManual = false;
	
end;


function mrGearboxMogli:doGearShift()

	--print(self.time .. " mrGearboxMogli:doGearShift - currentGear=" .. tostring(self.mrGbMCurrentGear));

	--only the server should call this
	
	
	self.realClutchEngaged = true;	
	
	local gearMaxSpeed = self.mrGbMGears[self.mrGbMCurrentGear].speed * self.mrGbMRanges[self.mrGbMCurrentRange].ratio
	if self.mrGbMNeutralActive then
		self.realShuttleDirection  = 0;
		self.realClutchEngaged     = false;
		self.realParkbrakeIsManual = false;
	else
		if self.mrGbMReverseActive then	
			self.realShuttleDirection = -1;
			gearMaxSpeed = gearMaxSpeed * self.mrGbMReverseRatio;
		else
			self.realShuttleDirection = 1;
		end
	end
	--local gearMinSpeedForMaxPower = gearMaxSpeed * self.mrGbMMinRpmAtMaxPower / self.mrGbMMaxRpm;--m/s			
	--print(self.time .. " mrGearboxMogli:doGearShift - "..string.format("mrGbMMinMaxRpmRatio=%1.2f", self.mrGbMMinMaxRpmRatio) .. string.format(" gearMinSpeedForMaxPower=%1.2f",gearMinSpeedForMaxPower)..string.format(" gearMaxSpeed=%1.2f", gearMaxSpeed));
	
	self:setManualGear3(true, gearMaxSpeed);	
	
end;



mrGearboxMogliSetGearEvent = {};
mrGearboxMogliSetGearEvent_mt = Class(mrGearboxMogliSetGearEvent, Event);

InitEventClass(mrGearboxMogliSetGearEvent, "mrGearboxMogliSetGearEvent");

 
function mrGearboxMogliSetGearEvent:emptyNew()
	local self = Event:new(mrGearboxMogliSetGearEvent_mt);
	self.className="mrGearboxMogliSetGearEvent";
	return self;
end;

function mrGearboxMogliSetGearEvent:new(object, gearNum, range, reverseActive, neutralActive)
	local self = mrGearboxMogliSetGearEvent:emptyNew();
	self.gearNum = gearNum;
	self.range = range;
	self.reverseActive = reverseActive;
	self.neutralActive = neutralActive;
	self.object = object;
	return self;
end;

function mrGearboxMogliSetGearEvent:readStream(streamId, connection)
	--both clients and server can receive this event
	local id = streamReadInt32(streamId);
	self.gearNum = streamReadInt32(streamId);
	self.range = streamReadInt32(streamId);
	self.reverseActive = streamReadBool(streamId);
	self.neutralActive = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;

function mrGearboxMogliSetGearEvent:writeStream(streamId, connection)
	--both clients and server can send this event
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt32(streamId, self.gearNum);
	streamWriteInt32(streamId, self.range);
	streamWriteBool(streamId, self.reverseActive);
	streamWriteBool(streamId, self.neutralActive);
end;

function mrGearboxMogliSetGearEvent:run(connection)
	----both clients and server can "run" this event (after reading it)	
	
	if not connection:getIsServer() then
		--we are the server, and so, we need to broadcast the new gearnum to all the other clients
		self.object:setGear(self.gearNum, self.range, self.reverseActive, false);	
	else
		--we are the client, and so we just receive the new gearnum
		self.object:setGear(self.gearNum, self.range, self.reverseActive, self.neutralActive, true);	
	end;
end;



mrGearboxMogliSetNextGearEvent = {};
mrGearboxMogliSetNextGearEvent_mt = Class(mrGearboxMogliSetNextGearEvent, Event);

InitEventClass(mrGearboxMogliSetNextGearEvent, "mrGearboxMogliSetNextGearEvent");

 
function mrGearboxMogliSetNextGearEvent:emptyNew()
	local self = Event:new(mrGearboxMogliSetNextGearEvent_mt);
	self.className="mrGearboxMogliSetNextGearEvent";
	return self;
end;

function mrGearboxMogliSetNextGearEvent:new(object, gearDown)
	local self = mrGearboxMogliSetNextGearEvent:emptyNew();
	self.gearDown = gearDown;
	self.object = object;
	return self;
end;

function mrGearboxMogliSetNextGearEvent:readStream(streamId, connection)
	--only the server receive the event
	local id = streamReadInt32(streamId);
	self.gearDown = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;

function mrGearboxMogliSetNextGearEvent:writeStream(streamId, connection)
	--only the clients send this event
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.gearDown);
end;

function mrGearboxMogliSetNextGearEvent:run(connection)
	--only the server "run" this event (after reading it)	
	self.object:setNextGear(self.gearDown, true);	
end;



mrGearboxMogliSetNextRangeEvent = {};
mrGearboxMogliSetNextRangeEvent_mt = Class(mrGearboxMogliSetNextRangeEvent, Event);

InitEventClass(mrGearboxMogliSetNextRangeEvent, "mrGearboxMogliSetNextRangeEvent");

 
function mrGearboxMogliSetNextRangeEvent:emptyNew()
	local self = Event:new(mrGearboxMogliSetNextRangeEvent_mt);
	self.className="mrGearboxMogliSetNextRangeEvent";
	return self;
end;

function mrGearboxMogliSetNextRangeEvent:new(object, rangeDown)
	local self = mrGearboxMogliSetNextRangeEvent:emptyNew();
	self.rangeDown = rangeDown;
	self.object = object;
	return self;
end;

function mrGearboxMogliSetNextRangeEvent:readStream(streamId, connection)
	--only the server receive the event
	local id = streamReadInt32(streamId);
	self.rangeDown = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;

function mrGearboxMogliSetNextRangeEvent:writeStream(streamId, connection)
	--only the clients send this event
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.rangeDown);
end;

function mrGearboxMogliSetNextRangeEvent:run(connection)
	--only the server "run" this event (after reading it)	
	self.object:setNextRange(self.rangeDown, true);	
end;



mrGearboxMogliClutchEvent = {};
mrGearboxMogliClutchEvent_mt = Class(mrGearboxMogliClutchEvent, Event);

InitEventClass(mrGearboxMogliClutchEvent, "mrGearboxMogliClutchEvent");

 
function mrGearboxMogliClutchEvent:emptyNew()
	local self = Event:new(mrGearboxMogliClutchEvent_mt);
	self.className="mrGearboxMogliClutchEvent";
	return self;
end;

function mrGearboxMogliClutchEvent:new(object,pressed)
	local self = mrGearboxMogliClutchEvent:emptyNew();
	self.object  = object;
	self.pressed = pressed;
	return self;
end;

function mrGearboxMogliClutchEvent:readStream(streamId, connection)
	--only the server receive the event
	local id = streamReadInt32(streamId);
	self.pressed = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;

function mrGearboxMogliClutchEvent:writeStream(streamId, connection)
	--only the clients send this event
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.pressed);
end;

function mrGearboxMogliClutchEvent:run(connection)
	--only the server "run" this event (after reading it)	
	self.object:clutch(pressed,true);	
end;